/*
*	name: formparse
*	function: decode a form-data and store the result in the file
*	usage:
*		formparse [-Er] [-o dir] [query] 
*		formparse -r -o dir query   # form-data is given by query
*		or
*		formparse -r -o dir	# form-data is given by stdin
*		---
*		option -E: simplly echo message body for debug
*		option -r: removes CR codes from the data
*
*	result:
*		decoded data is stored in files. the direcory structure is:
*		name/N	where N is 0, 1, ...
*		the count of duplicated name is in
*		name/clone
*
*	example1: in case of get-method
*		{ramfs >[2=1]} >/dev/null
*		formparse -r -o /tmp $QUERY_STRING
*
*	example2: in case of post-method
*		{ramfs >[2=1]} >/dev/null
*		formparse -r -o /tmp
*
*	note:
*		you will find some codes for multipart/form-data,
*		however they are still under construction.
*
*	auther: Kenji Arisawa <arisawa@aichi-u.ac.jp>
*/

#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>

#define NAMELEN 128

enum{
	EOD	= -1,
	NL = '\n',
};

static void error(char *fmt,...);
static int x2i(int i);
static char *qparse(char *qs);
static int cparse(char *b);
static void mparse(char *b);
static void xparse(char *dir);
static char *skipto(char *s, char *p);
static char *trim(char *s, char *p);
static char *value(char *s, char *name);
static Biobuf *Bnew1(char *name);
static Biobuf *Bnew2(char *dir, char *name);
static int readto(int fd, char *bp, int size, char del);

static int writefile(char *fn, char *b, int n);
static int readfile(char *fn, char *b, int n);
shapeup= 0;

char *odir="/tmp";

static void
usage(void)
{
	fprint(2,"usage: formparse [ -r] [-o dir] [query]\n");
	exits("usage");
}

static void
error(char *fmt,...)
{
	char buf[1024];
	va_list arg;
	va_start(arg, fmt);
	vseprint(buf, buf+sizeof(buf), fmt, arg);
	va_end(arg);
	print("HTTP/1.0\nContent-type: text/html\n\n");
	print("<html><head><title>Form processing error</title></head>");
	print("<body><h1>Error</h1><b>%s</b></body></html>\n", buf);
	exits("error");
}

void
main(int argc, char *argv[])
{
	char *qs;
	char *s, *t, *p;
	char b[80];
	int echoflag = 0;
	int n;

	ARGBEGIN{
	case 'o':
		odir=ARGF();
		break;
	case 'r':
		shapeup= 1;
		break;
	case 'E': // simply echo message body for debug
		echoflag=1;
		break;
	default:
		usage();
	}ARGEND
	if(chdir(odir) < 0)
		error("could not chdir to %s",odir);
	qs = nil;
	if(argv[0])
		qs = argv[0];
	if(qs)// parse x-www-form-urlencoded string
		while((qs = qparse(qs)));
	else{
		 s = getenv("HTTP_CONTENT_LENGTH");
		if(s == nil)
			error("no content length");
		n = atoi(s);
		if(echoflag){
			print("content-length=%d\n", n);
			s = malloc(n+1);
			n = read(0, s, n);
			print("read-length=%d\n", n);
			write(1, s, n);
			exits(nil);
		}
		t = getenv("HTTP_CONTENT_TYPE");
		if(t == nil)
			error("no content type");
		p = skipto(t,";");
		if(*p == ';')
			*p++ = 0;
		t = trim(t, " ");
		if(strcmp(t,"application/x-www-form-urlencoded") == 0){
			qs = malloc(n + 1);
			n = readn(0, qs, n);
			qs[n] = 0;
			while((qs = qparse(qs)));
		}
		else if(strcmp(t,"multipart/form-data") == 0){
			p = value(p, "boundary");
			snprint(b, sizeof b, "--%s", p);
			mparse(b);
		}
		else
			error("unknown format: %s", t);
	}
}

/*	conver hex char to int	*/
static int
x2i(int i)
{
	i=toupper(i);
	i = i - '0';
	if(i > 9) i = i - 'A' + '9' + 1;
	return i;
}

static int
readfile(char *fn, char *b, int n)
{
	int fd;
	fd = open(fn, OREAD);
	if(fd < 0)
		return -1;
	n = read(fd, b, n);
	b[n] = 0;
	close(fd);
	return n;
}

static int
writefile(char *fn, char *b, int n)
{
	int fd;
	fd = open(fn, OWRITE);
	if(fd < 0)
		return -1;
	n = write(fd, b, n);
	close(fd);
	return n;
}

static int
mkdir(char *name)
{
	Dir *dir;
	int fd;
	if((dir = dirstat(name)) == nil){ /* the name is first apeared */
		fd = create(name, OREAD, 0750 | DMDIR);
		if(fd < 0) /* somthing wrong */
			error("could not create %s", name);
		close(fd);
		return 0;
	}
	free(dir);
	return -1;
}


static Biobuf *
Bnew1(char *name)
{
	Biobuf *bout;
	char buf[20], *fn;
	int fd, n, fnsize;
	/*
	*	we must count the occurrence of the name.
	*	count it in:
	*		name/clone
	*/
	fnsize = strlen(name)+10;
	fn = malloc(fnsize);
	if(mkdir(name) == 0){
		snprint(fn, fnsize,"%s/clone",name);
		fd = create(fn, OWRITE, 0640);
		if(fd < 0) /* somthing wrong */
			error("%s not open", fn);
		write(fd,"0",1);
		close(fd);
	}
	/*
	*	read clone. increase the value. write it back
	*	the name and the value is sored in:
	*	name/0, name/1, ...
	*/
	if(readfile(fn, buf, sizeof(buf))<0)
		error("cannot read %s", fn);

	snprint(fn, fnsize,"%s/%s", name, buf);
	bout = Bopen(fn, OWRITE);
	if(!bout)
		error("%s not open", fn);
	n = atoi(buf) + 1;
	snprint(buf,sizeof(buf),"%d", n);
	snprint(fn, fnsize,"%s/clone",name);
	if(writefile(fn, buf, strlen(buf)) < 0)
		error("cannot write to %s", fn);
	free(fn);
	return bout;
}

static Biobuf *
Bnew2(char *dir, char *name)
{
	Biobuf *b;
	char *fn;
	int fnsize;
	mkdir(dir);
	fnsize = strlen(dir) + strlen(name) + 10;
	fn = malloc(fnsize);
	snprint(fn, fnsize,"%s/%s",dir, name);
	b = Bopen(fn, OWRITE);
	if(b == nil) /* somthing wrong */
		error("%s not open", fn);
	free(fn);
	return b;
}

/*
*	parse single name and the value
*	return:
*	0:	End of data or fatal error
*	else:	next point to parse; parsing single name/value pair has finished
*/
static char *
qparse(char *qs)
{
	char *s, *s0;
	char name[NAMELEN];
	int c, c1;
	Biobuf *bout;

	/*
	*	regular format is,
	*	name=val&name=val&...name=val
	*	but the data can be forged, so we must be careful.
	*	example of forged patterns:
	*	&&&&
	*	====
	*	=&=&=
	*/

	/*
	*	parse a name
	*	we restrict chars in the name to be alnum
	*/
	s = s0 = name;
	while((c = *qs++)){
		if(s - s0 == NAMELEN - 1){ /* truncate here */
			break;
		}
		if(c == '='){
			break;
		}
		if(isalnum(c)){
			*s++ = c;
			continue;
		}
		/* something wrong. skip it */
	}
	*s = 0;
	if(c == 0)
		return 0;

	/* sweep out extra chars */
	while(c != '=' && (c = *qs++));

	/* check the name length */
	if(strlen(name) == 0){ /*something wrong. skip to next name */
		while((c = *qs++) && c != '&');
		return qs;
	}

	bout = Bnew1(name);

	/* get the value */
	while((c = *qs++)){
		switch(c){
		case '&':
			Bterm(bout);
			return qs;
		case '+':
			c = ' ';
			break;
		case '%':
			c = *qs++;
			if(c == 0) goto err;
			c1 = c;
			c = *qs++;
			if(c == 0) goto err;
			c = 16*x2i(c1) + x2i(c);
			if(c == '\r' && shapeup) continue;
			break;
		default:
			;
		}
		Bputc(bout, c);
	}
err:
	Bterm(bout);
	return 0;
}


/*	parse multipart/form-data	*/
static void
mparse(char *b)
{
	char buf[256];
	int n,m,bsize;
	m = sizeof buf;
	bsize = strlen(b);

	n = readto(0,buf, m, NL);
	if(buf[n-1] != NL)
		error("first boundary too long: len=%d", n);
	buf[n-1] = 0;
	if(strncmp(buf, b, bsize) != 0)
		error("not start with boundary");
	while(cparse(b));
}

/*	parse Content-Disposition	*/
static  int
cparse(char *b)
{
	char buf[4096];
	char *argv[3];
	char *s, *t;
	int n,m,bsize;
	Biobuf *bout;
	bsize = strlen(b);
	m = sizeof buf;
	n = readto(0,buf, m, NL);
	if(buf[n-1] != NL)
		error("first line too long in content block");
	buf[n-1] = 0;
	s = trim(buf," \r");
	t = skipto(s, ":");
	if(*t)
		*t++ = 0;
	s = trim(s, " ");
	if(strcmp(s, "Content-Disposition") != 0)
		error("error: [%s] should be Content-Disposition", s);
	n = getfields(t, argv, 3, 0, ";");
	if(n < 2)
		error("Content-Disposiotion too short");
	s = trim(argv[0], " ");
	if(strcmp(s, "form-data") != 0)
		error("error: [%s] should be form-data", s);
	s = value(argv[1], "name");
	if(s == nil)
		error("no name");
	bout = nil;
	switch(n){
	case 2:
		bout = Bnew1(s);
		break;
	case 3:
		t = value(argv[2], "filename");
		bout = Bnew2(s, t);
		break;
	default:
		error("too many fileds");
	}
	n = readto(0,buf, m, NL);
	if(n != 2 || buf[0] != '\r' || buf[1] != NL)
		error("not a empty line: %d [%s]", n,buf);
	while((n = readto(0, buf, m, NL)) > 0 && strncmp(buf, b, bsize) != 0)
		Bwrite(bout, buf, n);
	Bterm(bout);


	if(strncmp(buf + bsize, "--\r\n", 5) == 0)
		return 0; // finished
	if(strncmp(buf + bsize, "\r\n", 2) == 0)
		return 1; // continue
	return 0; // error
}


static char *
skip(char *s, char *p)
{
	while(*s && strchr(p, *s))
		s++;
	return s;
}

static char *
skipto(char *s, char *p)
{
	while(*s){
		if(strchr(p, *s))
			break;
		s++;
	}
	return s;
}

static char*
trim(char *s, char *p)
{
	char *t;
	while(strchr(p,*s))
		s++;
	t = strchr(s, 0);
	t--;
	while(strchr(p,*t ))
		t--;
	t++;
	*t = 0;
	return s;
}

static char *
value(char *s, char *name)
{
	char *t;
	t = skipto(s, "=");
	if(*t == 0)
		error("syntax error");
	*t++ = 0;
	s = trim(s, " ");
	if(strcmp(s, name) != 0)
		error("error in name %s", s);
	return trim(t, " \"");
}


/*
*	readto
*	usage: n = readto(fd, buf, bufsize, DEL)
*	Readto is just like read, but read chars until DEL (delimiter)
*	and returns the number of chars read.
*	If DEL is read in buf then:
*		n <= bufsize
*	and 
*		buf[n-1] == DEL.
*	If a line is larger than bufsize then:
*		n == bufsize
*	and
*		buf[n-1] != DEL.
*
*	-kenar-
*/
static int
readto(int fd, char *bp, int size, char del)
{
	static char buf[4*4096];
	static char *s;
	static int n=0;
	int k=0;
	int i;
	USED(bp);
	USED(size);
	for(;;){
		if(n == 0){
			n = read(fd, buf, sizeof(buf));
			s = buf;
			if(n <= 0){
				n = 0;
				if(k == 0) return -1;
				else return k;
			}
		}
		for (i=0; i < n &&  i < size && *s != del;  i++) *bp++ = *s++;
		n -= i;
		size -= i;
		k += i;
		/*
		*	here we have three cases:
		*	1.  size == 0		-> simply return
		*	2.  n == 0			-> reread
		*	3.  *s == '\n'		-> *bp++ = *s++
		*	keep this order in evaluation.
		*/
		if(size==0) break;
		if(n == 0) continue;
		/* then *s must be del and also n>0 && size>0 */
		*bp++ = *s++; n--; size--; k++;
		break;
	}
	return k;
}
